<?php

namespace Drupal\commerce_klarna_checkout;

use Drupal\commerce_klarna_checkout\Event\KlarnaCheckoutEvents;
use Drupal\commerce_klarna_checkout\Event\OrderRequestEvent;
use Drupal\commerce_order\AdjustmentTransformerInterface;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\Price;
use Drupal\commerce_product\Entity\ProductVariationInterface;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Url;
use Drupal\profile\Entity\ProfileInterface;
use Klarna\Rest\Checkout\Order as KlarnaOrder;
use Klarna\Rest\OrderManagement\Order as KlarnaOrderManagement;
use Klarna\Rest\Transport\ConnectorInterface;
use Klarna\Rest\Transport\GuzzleConnector;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

class KlarnaManager implements KlarnaManagerInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The adjustment transformer.
   *
   * @var \Drupal\commerce_order\AdjustmentTransformerInterface
   */
  protected $adjustmentTransformer;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The payment gateway plugin configuration.
   *
   * @var array
   */
  protected $config;

  /**
   * The Klarna HTTP transport connector.
   *
   * @var \Klarna\Rest\Transport\ConnectorInterface
   */
  protected $connector;

  /**
   * Constructs a new KlarnaManager object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_order\AdjustmentTransformerInterface $adjustment_transformer
   *   The adjustment transformer.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param array $config
   *   The payment gateway plugin configuration array.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AdjustmentTransformerInterface $adjustment_transformer, EventDispatcherInterface $event_dispatcher, array $config) {
    $this->entityTypeManager = $entity_type_manager;
    $this->adjustmentTransformer = $adjustment_transformer;
    $this->eventDispatcher = $event_dispatcher;
    $this->config = $config;
    $this->initConnector();
  }

  /**
   * Initialize the HTTP transport connector.
   */
  protected function initConnector() {
    // See https://developers.klarna.com/api/#api-urls.
    if ($this->config['purchase_country'] === 'US') {
      $endpoint = $this->config['mode'] === 'test' ? ConnectorInterface::NA_TEST_BASE_URL : ConnectorInterface::NA_BASE_URL;
    }
    else {
      $endpoint = $this->config['mode'] === 'test' ? ConnectorInterface::EU_TEST_BASE_URL : ConnectorInterface::EU_BASE_URL;
    }
    $this->connector = GuzzleConnector::create(
      $this->config['username'],
      $this->config['password'],
      $endpoint
    );
  }

  /**
   * {@inheritdoc}
   */
  public function acknowledgeOrder(OrderInterface $order) {
    $this->assertOrder($order);
    $klarna_order = new KlarnaOrderManagement($this->connector, $order->getData('klarna_order_id'));
    $klarna_order->acknowledge();
  }

  /**
   * {@inheritdoc}
   */
  public function cancelOrder(OrderInterface $order) {
    $this->assertOrder($order);
    $klarna_order = new KlarnaOrderManagement($this->connector, $order->getData('klarna_order_id'));
    $klarna_order->cancel();
  }

  /**
   * {@inheritdoc}
   */
  public function createOrder(OrderInterface $order, array $merchant_urls) {
    $request_data = $this->buildOrderRequest($order, $merchant_urls);
    $event = new OrderRequestEvent($order, $request_data);
    $this->eventDispatcher->dispatch(KlarnaCheckoutEvents::CREATE_ORDER_REQUEST, $event);
    $klarna_order = new KlarnaOrder($this->connector);
    $klarna_order = $klarna_order->create($event->getRequestData());
    $klarna_order->fetch();
    return $klarna_order;
  }

  /**
   * {@inheritdoc}
   */
  public function getOrder(OrderInterface $order) {
    $this->assertOrder($order);
    $klarna_order = new KlarnaOrder($this->connector, $order->getData('klarna_order_id'));
    $klarna_order->fetch();
    return $klarna_order;
  }

  /**
   * {@inheritdoc}
   */
  public function loadOrder($klarna_order_id) {
    $klarna_order = new KlarnaOrder($this->connector, $klarna_order_id);
    $klarna_order->fetch();
    return $klarna_order;
  }

  /**
   * {@inheritdoc}
   */
  public function updateOrder(OrderInterface $order, array $merchant_urls) {
    $request_data = $this->buildOrderRequest($order, $merchant_urls);
    $event = new OrderRequestEvent($order, $request_data);
    $this->eventDispatcher->dispatch(KlarnaCheckoutEvents::UPDATE_ORDER_REQUEST, $event);
    $klarna_order = new KlarnaOrder($this->connector, $order->getData('klarna_order_id'));
    $klarna_order->update($event->getRequestData());
    return $klarna_order;
  }

  /**
   * {@inheritdoc}
   */
  public function createCapture(OrderInterface $order, array $data) {
    $this->assertOrder($order);
    if (!isset($data['captured_amount'])) {
      throw new \InvalidArgumentException("Missing required 'captured_amount' key in the passed data.");
    }
    $data += [
      'order_lines' => $this->buildOrderLines($order),
    ];
    $klarna_order = new KlarnaOrderManagement($this->connector, $order->getData('klarna_order_id'));
    return $klarna_order->createCapture($data);
  }

  /**
   * {@inheritdoc}
   */
  public function createRefund(OrderInterface $order, array $data) {
    $this->assertOrder($order);
    if (!isset($data['refunded_amount'])) {
      throw new \InvalidArgumentException("Missing required 'captured_amount' key in the passed data.");
    }
    $klarna_order = new KlarnaOrderManagement($this->connector, $order->getData('klarna_order_id'));
    return $klarna_order->refund($data);
  }

  /**
   * {@inheritdoc}
   */
  public function updateAuthorization(OrderInterface $order, array $data = []) {
    $this->assertOrder($order);
    $data += [
      'order_amount' => $this->toMinorUnits($order->getTotalPrice()),
      'order_lines' => $this->buildOrderLines($order),
    ];
    $klarna_order = new KlarnaOrderManagement($this->connector, $order->getData('klarna_order_id'));
    return $klarna_order->updateAuthorization($data);
  }

  /**
   * Builds the order request array for the given order.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   * @param array $merchant_urls
   *   The merchant URLS.
   *
   * @return array
   *   The order request array.
   */
  protected function buildOrderRequest(OrderInterface $order, array $merchant_urls) {
    foreach (['checkout', 'confirmation', 'push'] as $required_key) {
      if (!isset($merchant_urls[$required_key])) {
        throw new \InvalidArgumentException(sprintf('Missing required key %s in the provided $merchant_urls array.', $required_key));
      }
    }
    if (is_null($order->getTotalPrice())) {
      throw new \InvalidArgumentException(sprintf('Cannot build the order request with an empty order total for order id %s.', $order->id()));
    }
    $profiles = $order->collectProfiles();
    $adjustments = $order->collectAdjustments();
    $terms_path = $this->config['terms_path'];
    $terms_uri = UrlHelper::isExternal($terms_path) ? Url::fromUri($terms_path) : Url::fromUserInput('/' . ltrim($terms_path, '/'), ['absolute' => TRUE]);
    $data = [
      'purchase_country' => $this->config['purchase_country'],
      'purchase_currency' => $order->getTotalPrice()->getCurrencyCode(),
      'name' => $order->getStore()->label(),
      'locale' => $this->config['locale'],
      'order_amount' => $this->toMinorUnits($order->getTotalPrice()),
      'merchant_urls' => [
        'terms' => $terms_uri->toString(),
        'checkout' => $merchant_urls['checkout'],
        'confirmation' => $merchant_urls['confirmation'],
        'push' => $merchant_urls['push'],
      ],
      'merchant_reference1' => $order->getOrderNumber() ?: $order->id(),
      'merchant_reference2' => $order->id(),
      'options' => [
        'allow_separate_shipping_address' => (bool) $this->config['allow_separate_shipping_address'],
      ],
      'order_lines' => $this->buildOrderLines($order),
    ];

    if (!empty($this->config['allowed_customer_types'])) {
      $data['options']['allowed_customer_types'] = $this->config['allowed_customer_types'];
    }

    // Send the billing profile only if not null.
    if (isset($profiles['billing'])) {
      $billing_address = $this->buildAddress($profiles['billing']);
      $data['billing_address'] = $billing_address;

      if ($order->getEmail()) {
        $data['billing_address']['email'] = $order->getEmail();
      }
    }

    $tax_total = $this->getAdjustmentsTotal($adjustments, ['tax']);
    $data['order_tax_amount'] = $tax_total ? $this->toMinorUnits($tax_total) : 0;
    return $data;
  }

  /**
   * Builds the order lines for Klarna.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   *
   * @return array
   *   The order lines as expected by Klarna.
   */
  protected function buildOrderLines(OrderInterface $order) {
    $order_lines = [];
    foreach ($order->getItems() as $order_item) {
      // Fallback to the order item ID.
      $reference = $order_item->id();
      $purchased_entity = $order_item->getPurchasedEntity();

      // Send the "SKU" as the "reference" for product variations.
      if ($purchased_entity instanceof ProductVariationInterface) {
        $reference = $purchased_entity->getSku();
      }

      $order_line = [
        // The reference is limited to 64 characters.
        'reference' => mb_substr($reference, 0, 64),
        'name' => $order_item->label(),
        'quantity' => (int) $order_item->getQuantity(),
        'unit_price' => $this->toMinorUnits($order_item->getUnitPrice()),
        'total_amount' => $this->toMinorUnits($order_item->getAdjustedTotalPrice(['promotion'])),
        'tax_rate' => 0,
        'total_tax_amount' => 0,
      ];

      // Only pass included tax adjustments (i.e "VAT"), non included tax
      // adjustments are passed separately (i.e "Sales tax").
      $tax_adjustments = $order_item->getAdjustments(['tax']);
      if ($tax_adjustments && $tax_adjustments[0]->isIncluded()) {
        $tax_rate = $tax_adjustments[0]->getPercentage();
        $order_line = array_merge($order_line, [
          'tax_rate' => (int) Calculator::multiply($tax_rate, '10000'),
          'total_tax_amount' => $this->toMinorUnits($tax_adjustments[0]->getAmount()),
        ]);
      }

      // Check if we have promotion adjustments.
      $promotion_adjustments = $order_item->getAdjustments(['promotion']);
      if ($promotion_adjustments) {
        $promotions_total = $this->getAdjustmentsTotal($promotion_adjustments);
        $order_line['total_discount_amount'] = $this->toMinorUnits($promotions_total->multiply('-1'));
      }
      $order_lines[] = $order_line;
    }
    $adjustment_types_mapping = [
      'tax' => 'sales_tax',
      'fee' => 'surcharge',
    ];
    // Shipping is handled separately.
    $exclude_adjustment_types = ['shipping', 'promotion', 'shipping_promotion'];
    foreach ($order->collectAdjustments() as $adjustment) {
      $adjustment_type = $adjustment->getType();
      // Skip included adjustments and the ones we don't handle.
      if ($adjustment->isIncluded() ||
        in_array($adjustment_type, $exclude_adjustment_types)) {
        continue;
      }
      $order_line = [
        // The reference is limited to 64 characters.
        'reference' => $adjustment->getSourceId() ? mb_substr($adjustment->getSourceId(), 0, 64) : '',
        'name' => $adjustment->getLabel(),
        'quantity' => 1,
        'tax_rate' => 0,
        'total_tax_amount' => 0,
        'unit_price' => $this->toMinorUnits($adjustment->getAmount()),
        'total_amount' => $this->toMinorUnits($adjustment->getAmount()),
      ];

      // Map the adjustment type to the type expected by Klarna.
      if (isset($adjustment_types_mapping[$adjustment_type])) {
        $order_line['type'] = $adjustment_types_mapping[$adjustment_type];
      }

      $order_lines[] = $order_line;
    }
    if ($order->hasField('shipments') && !$order->get('shipments')->isEmpty()) {
      /** @var \Drupal\commerce_shipping\Entity\ShipmentInterface[] $shipments */
      $shipments = $order->get('shipments')->referencedEntities();
      foreach ($shipments as $shipment) {
        $shipment_amount = $shipment->getAmount();
        if (empty($shipment_amount)) {
          continue;
        }
        // Check if there are included tax adjustments.
        $tax_adjustments = array_filter($shipment->getAdjustments(['tax']), function($adjustment) {
          return $adjustment->isIncluded();
        });
        $shipping_line = [
          'name' => (string) t('Shipping'),
          'quantity' => 1,
          'type' => 'shipping_fee',
        ];
        // If there are no included tax adjustments, there's no need to split
        // the shipment amounts across multiple VAT rates.
        if (!$tax_adjustments) {
          $shipping_line += [
            'total_amount' => $this->toMinorUnits($shipment->getAdjustedAmount(['shipping_promotion'])),
            'tax_rate' => 0,
            'total_tax_amount' => 0,
          ];
          // The unit price shouldn't include discounts, we cannot use the
          // original amount as the "unit_price" because it won't reflect the
          // correct price if shipping rates were altered.
          $unit_price = $shipment->getAmount();
          $promotion_adjustments = $shipment->getAdjustments(['shipping_promotion']);
          if ($promotion_adjustments) {
            $promotions_total = $this->getAdjustmentsTotal($promotion_adjustments);
            $shipping_line['total_discount_amount'] = $this->toMinorUnits($promotions_total->multiply('-1'));

            // We need to add any included promotion adjustment to the unit
            // price.
            foreach ($promotion_adjustments as $adjustment) {
              if (!$adjustment->isIncluded()) {
                continue;
              }
              $unit_price->add($adjustment->getAmount()->multiply('-1'));
            }
          }
          $shipping_line['unit_price'] = $this->toMinorUnits($unit_price);
          $order_lines[] = $shipping_line;
        }
        else {
          /** @var \Drupal\commerce_klarna_checkout\ShipmentPriceSplitterInterface $shipment_price_splitter */
          $shipment_price_splitter = \Drupal::service('commerce_klarna_checkout.shipment_price_splitter');
          $shipment_amounts = $shipment_price_splitter->split($order, $shipment);
          foreach ($shipment_amounts as $amounts) {
            $order_lines[] = $shipping_line + [
              'unit_price' => $this->toMinorUnits($amounts['unit_amount']),
              'total_amount' => $this->toMinorUnits($amounts['adjusted_shipment_amount']),
              'tax_rate' => (int) Calculator::multiply($amounts['tax_rate'], '10000'),
              'total_tax_amount' => $this->toMinorUnits($amounts['tax_amount']),
              'total_discount_amount' => $this->toMinorUnits($amounts['promotions_amount']),
            ];
          }
        }
      }
    }

    return $order_lines;
  }

  /**
   * Builds an address in a format expected by Klarna for the given profile.
   *
   * @param \Drupal\profile\Entity\ProfileInterface $profile
   *   The profile to build an address for.
   *
   * @return array
   *   The formatted address for use in Klarna API calls.
   *
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   *   If the complex data structure is unset and no item can be created.
   */
  protected function buildAddress(ProfileInterface $profile) {
    /** @var \Drupal\address\AddressInterface $address */
    $address = $profile->get('address')->first();
    return [
      'organization_name' => $address->getOrganization(),
      'given_name' => $address->getGivenName(),
      'family_name' => $address->getFamilyName(),
      'country' => $address->getCountryCode(),
      'postal_code' => $address->getPostalCode(),
      'city' => $address->getLocality(),
      'region' => $address->getAdministrativeArea(),
      'street_address' => $address->getAddressLine1(),
      'street_address2' => $address->getAddressLine2(),
    ];
  }

  /**
   * Calculates the total for the given adjustments.
   *
   * @param \Drupal\commerce_order\Adjustment[] $adjustments
   *   The adjustments.
   * @param string[] $adjustment_types
   *   The adjustment types to include in the calculation.
   *   Examples: fee, promotion, tax. Defaults to all adjustment types.
   *
   * @return \Drupal\commerce_price\Price|null
   *   The adjustments total, or NULL if no matching adjustments were found.
   */
  protected function getAdjustmentsTotal(array $adjustments, array $adjustment_types = []) {
    $adjustments_total = NULL;
    $matching_adjustments = [];

    foreach ($adjustments as $adjustment) {
      if ($adjustment_types && !in_array($adjustment->getType(), $adjustment_types)) {
        continue;
      }
      $matching_adjustments[] = $adjustment;
    }
    if ($matching_adjustments) {
      $matching_adjustments = $this->adjustmentTransformer->processAdjustments($matching_adjustments);
      foreach ($matching_adjustments as $adjustment) {
        $adjustments_total = $adjustments_total ? $adjustments_total->add($adjustment->getAmount()) : $adjustment->getAmount();
      }
    }

    return $adjustments_total;
  }

  /**
   * Converts the given amount to its minor units.
   *
   * For example, 9.99 USD becomes 999 (Copied from PaymentGatewayBase).
   *
   * @param \Drupal\commerce_price\Price $amount
   *   The amount.
   *
   * @return int
   *   The amount in minor units, as an integer.
   */
  protected function toMinorUnits(Price $amount) {
    $currency_storage = $this->entityTypeManager->getStorage('commerce_currency');
    /** @var \Drupal\commerce_price\Entity\CurrencyInterface $currency */
    $currency = $currency_storage->load($amount->getCurrencyCode());
    $fraction_digits = $currency->getFractionDigits();
    $number = $amount->getNumber();
    if ($fraction_digits > 0) {
      $number = Calculator::multiply($number, pow(10, $fraction_digits));
    }

    return (int) round($number, 0);
  }

  /**
   * Asserts that the order holds a reference to the Klarna order ID.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   *
   * @throws \InvalidArgumentException
   *   Thrown if the order doesn't hold a reference to the Klarna order ID.
   */
  protected function assertOrder(OrderInterface $order) {
    if (!$order->getData('klarna_order_id')) {
      throw new \InvalidArgumentException(sprintf('Missing Klarna order ID for order %s.', $order->id()));
    }
  }

}
