<?php

namespace Drupal\commerce_klarna_checkout\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_klarna_checkout\KlarnaManagerFactoryInterface;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\Price;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\profile\Entity\ProfileInterface;
use Klarna\Rest\Checkout\Order as KlarnaOrder;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the Klarna checkout payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "klarna_checkout",
 *   label = "Klarna Checkout",
 *   display_label = "Klarna Checkout",
 *    forms = {
 *     "offsite-payment" = "Drupal\commerce_klarna_checkout\PluginForm\OffsiteRedirect\KlarnaCheckoutForm",
 *   },
 *   requires_billing_information = FALSE,
 * )
 */
class KlarnaCheckout extends OffsitePaymentGatewayBase implements KlarnaCheckoutInterface {

  /**
   * The Klarna manager factory.
   *
   * @var \Drupal\commerce_klarna_checkout\KlarnaManagerFactoryInterface
   */
  protected $klarnaManagerFactory;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Constructs a new KlarnaCheckout object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \Drupal\commerce_klarna_checkout\KlarnaManagerFactoryInterface $klarna_manager_factory
   *   The Klarna manager factory.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, PaymentTypeManager $payment_type_manager, PaymentMethodTypeManager $payment_method_type_manager, TimeInterface $time, KlarnaManagerFactoryInterface $klarna_manager_factory, LoggerInterface $logger, ModuleHandlerInterface $module_handler) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);

    $this->klarnaManagerFactory = $klarna_manager_factory;
    $this->logger = $logger;
    $this->moduleHandler = $module_handler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('commerce_klarna_checkout.manager_factory'),
      $container->get('logger.channel.commerce_klarna_checkout'),
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'username' => '',
      'password' => '',
      'capture' => FALSE,
      'purchase_country' => '',
      'locale' => 'sv-se',
      'terms_path' => '',
      'update_billing_profile' => TRUE,
      'update_shipping_profile' => FALSE,
      'allowed_customer_types' => [],
      'allow_separate_shipping_address' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['username'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Username'),
      '#default_value' => $this->configuration['username'],
      '#required' => TRUE,
    ];

    $form['password'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Password'),
      '#default_value' => $this->configuration['password'],
      '#required' => TRUE,
    ];

    $form['capture'] = [
      '#type' => 'radios',
      '#title' => $this->t('Transaction mode'),
      '#options' => [
        TRUE => $this->t('Authorize and capture'),
        FALSE => $this->t('Authorize only (requires manual capture after checkout)'),
      ],
      '#default_value' => (int) $this->configuration['capture'],
    ];

    $form['purchase_country'] = [
      '#type' => 'select',
      '#options' => [
        'AT' => $this->t('Austria'),
        'AU' => $this->t('Australia'),
        'DK' => $this->t('Denmark'),
        'FI' => $this->t('Finland'),
        'DE' => $this->t('Germany'),
        'NL' => $this->t('Netherlands'),
        'NO' => $this->t('Norway'),
        'SE' => $this->t('Sweden'),
        'CH' => $this->t('Switzerland'),
        'GB' => $this->t('United Kingdom'),
        'US' => $this->t('United States'),
      ],
      '#title' => t('Purchase country'),
      '#default_value' => $this->configuration['purchase_country'],
      '#required' => TRUE,
    ];

    $form['locale'] = [
      '#type' => 'select',
      '#title' => $this->t('Locale'),
      '#default_value' => $this->configuration['locale'],
      '#required' => TRUE,
      '#options' => [
        'en-au' => $this->t('English (en-au)'),
        'en-at' => $this->t('English (en-at)'),
        'en-ch' => $this->t('English (en-ch)'),
        'en-de' => $this->t('English (en-de)'),
        'en-dk' => $this->t('English (en-dk)'),
        'en-us' => $this->t('English (en-us)'),
        'en-gb' => $this->t('English (en-gb)'),
        'en-fi' => $this->t('English (en-fi)'),
        'en-nl' => $this->t('English (en-nl)'),
        'en-no' => $this->t('English (en-no)'),
        'da-dk' => $this->t('Danish (da-dk)'),
        'nl-nl' => $this->t('Dutch (nl-nl)'),
        'fr-ch' => $this->t('French (fr-ch)'),
        'fi-fi' => $this->t('Finnish (fi-fi)'),
        'de-de' => $this->t('German (de-de)'),
        'it-ch' => $this->t('Italian (it-ch)'),
        'nb-no' => $this->t('Norwegian (nb-no)'),
        'sv-se' => $this->t('Swedish (sv-se)'),
        'sv-fi' => $this->t('Swedish (sv-fi)'),
      ],
    ];

    $form['terms_path'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Path to terms and conditions page'),
      '#default_value' => $this->configuration['terms_path'],
      '#required' => TRUE,
    ];

    $form['update_billing_profile'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Update the billing customer profile with address information the customer enters at Klarna.'),
      '#default_value' => $this->configuration['update_billing_profile'],
    ];

    $form['update_shipping_profile'] = [
      '#type' => 'checkbox',
      '#title' => t('Update the shipping customer profile (if present) with address information the customer enters at Klarna.'),
      '#default_value' => $this->configuration['update_shipping_profile'],
      '#access' => $this->moduleHandler->moduleExists('commerce_shipping'),
    ];

    $form['options'] = [
      '#type' => 'details',
      '#title' => $this->t('Options'),
      '#open' => TRUE,
    ];

    $form['options']['allowed_customer_types'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Allowed customer types'),
      '#options' => [
        'person' => $this->t('Person'),
        'organization' => $this->t('Organization'),
      ],
      '#default_value' => $this->configuration['allowed_customer_types'],
      '#parents' => array_merge($form['#parents'], ['allowed_customer_types']),
    ];

    $form['options']['allow_separate_shipping_address'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Allow separate shipping address'),
      '#description' => $this->t('If true, the consumer can enter different billing and shipping addresses.'),
      '#default_value' => $this->configuration['allow_separate_shipping_address'],
      '#parents' => array_merge($form['#parents'], ['allow_separate_shipping_address']),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    if ($form_state->getErrors()) {
      return;
    }
    $values = $form_state->getValue($form['#parents']);
    $this->configuration['username'] = $values['username'];
    $this->configuration['password'] = $values['password'];
    $this->configuration['capture'] = !empty($values['capture']);
    $this->configuration['purchase_country'] = $values['purchase_country'];
    $this->configuration['locale'] = $values['locale'];
    $this->configuration['terms_path'] = $values['terms_path'];
    $this->configuration['update_billing_profile'] = $values['update_billing_profile'];
    $customer_types = array_values(array_filter($values['allowed_customer_types']));
    $this->configuration['update_shipping_profile'] = !empty($values['update_shipping_profile']);

    if ($customer_types) {
      $this->configuration['allowed_customer_types'] = $customer_types;
    }
    $this->configuration['allow_separate_shipping_address'] = $values['allow_separate_shipping_address'];
  }

  /**
   * {@inheritdoc}
   */
  public function getNotifyUrl() {
    $notify_url = parent::getNotifyUrl();
    $notify_url->setOption('query', [
      'klarna_order_id' => '{checkout.order.id}',
    ]);
    return $notify_url;
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    // According to the Klarna API documentation, there should be an attempt to
    // acknowledge the order from the confirmation callback.
    $this->acknowledgeOrder($order);
  }

  /**
   * {@inheritdoc}
   */
  public function onNotify(Request $request) {
    $klarna_order_id = $request->query->get('klarna_order_id');

    if (empty($klarna_order_id)) {
      $this->logger->error('Cannot acknowledge the Klarna order: no order ID provided by Klarna.');
      return NULL;
    }
    /** @var \Drupal\commerce_payment\PaymentStorageInterface $payment_storage */
    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $payment = $payment_storage->loadByRemoteId($klarna_order_id);

    // If a payment already exists in the system for the given order, it means
    // it was created by onReturn().
    if ($payment) {
      return NULL;
    }

    try {
      $klarna_manager = $this->klarnaManagerFactory->get($this->getConfiguration());
      $klarna_order = $klarna_manager->loadOrder($klarna_order_id);
      // The order_id is passed to Klarna via the "merchant_reference2" field,
      // so use that for retrieving the order.
      /** @var \Drupal\commerce_order\Entity\OrderInterface $order */
      $order = $this->entityTypeManager->getStorage('commerce_order')->load($klarna_order['merchant_reference2']);

      if (!$order) {
        $this->logger->error('Cannot find a matching order for the given Klarna order ID #@klarna_order_id.', ['@klarna_order_id' => $klarna_order_id]);
        return NULL;
      }

      $this->acknowledgeOrder($order, $klarna_order);
    }
    catch (\Exception $exception) {
      $this->logger->error($exception->getMessage());
      return NULL;
    }
  }

  /**
   * Acknowledge the given Klarna order, and optionally capture the payment if
   * configured to do so.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   * @param \Klarna\Rest\Checkout\Order $klarna_order
   *   (optional) The Klarna order (fetched from Klarna if not specified).
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function acknowledgeOrder(OrderInterface $order, KlarnaOrder $klarna_order = NULL) {
    try {
      $klarna_manager = $this->klarnaManagerFactory->get($this->getConfiguration());
      $klarna_order = $klarna_order ?: $klarna_manager->getOrder($order);
      $klarna_manager->acknowledgeOrder($order);
    }
    catch (\Exception $exception) {
      $this->logger->error('An exception occurred while acknowledging the order #@order_id.', ['@order_id' => $order->id()]);
      throw new PaymentGatewayException($exception->getMessage());
    }
    $payment_amount = $this->fromMinorUnits($klarna_order['order_amount'], $klarna_order['purchase_currency']);
    /** @var \Drupal\commerce_payment\PaymentStorageInterface $payment_storage */
    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $payment = $payment_storage->create([
      'state' => 'authorization',
      // Use the amount we get from Klarna, rather than the order total, since
      // it might be different.
      'amount' => $payment_amount,
      'payment_gateway' => $this->parentEntity ? $this->parentEntity->id() : $this->entityId,
      'order_id' => $order->id(),
      'test' => $this->getMode() == 'test',
      'remote_id' => $klarna_order->getId(),
      'remote_state' => $klarna_order['status'],
    ]);
    $order_total = $order->getTotalPrice();

    // Should we throw an exception when this happens?
    if ($payment_amount->compareTo($order_total) !== 0) {
      $this->logger->notice('Order total mismatch: (Klarna total: @klarna_total, Order total: @order_total).', [
        '@klarna_total' => $payment_amount->getNumber(),
        '@order_total' => $order_total->getNumber(),
      ]);
    }

    // Update the billing profile if configured to do so.
    if (!empty($this->configuration['update_billing_profile']) && isset($klarna_order['billing_address'])) {
      $profile = $order->getBillingProfile();
      if (!$profile) {
        $profile = $this->entityTypeManager->getStorage('profile')->create([
          'uid' => 0,
          'type' => 'customer',
        ]);
        $order->setBillingProfile($profile);
      }
      $this->populateProfile($profile, $klarna_order['billing_address']);
      $profile->save();
    }

    // Update the shipping profile if configured to do so.
    if (!empty($this->configuration['update_shipping_profile']) && isset($klarna_order['shipping_address'])) {
      $profiles = $order->collectProfiles();

      if (isset($profiles['shipping'])) {
        $this->populateProfile($profiles['shipping'], $klarna_order['shipping_address']);
        $profiles['shipping']->save();
      }
    }

    if (isset($klarna_order['billing_address']['email'])) {
      $order->setEmail($klarna_order['billing_address']['email']);
    }

    // We have to save the order since the billing profile and/or the email were
    // potentially updated.
    $order->setData('klarna_order_acknowledged', TRUE);
    $order->setRefreshState(OrderInterface::REFRESH_SKIP);
    $order->save();

    // Capture the payment right away, if configured to do so.
    if (!empty($this->configuration['capture'])) {
      $this->capturePayment($payment);
    }
    else {
      $payment->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(PaymentInterface $payment, Price $amount = NULL) {
    $this->assertPaymentState($payment, ['authorization']);
    // If not specified, capture the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $klarna_manager = $this->klarnaManagerFactory->get($this->getConfiguration());

    try {
      $parameters = [
        'captured_amount' => $this->toMinorUnits($amount),
      ];
      $capture = $klarna_manager->createCapture($payment->getOrder(), $parameters);
      $payment->setState('completed');
      $payment->setAmount($amount);
      $payment->save();
    }
    catch (\Exception $exception) {
      throw new PaymentGatewayException($exception->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment) {
    $this->assertPaymentState($payment, ['authorization']);

    try {
      $klarna_manager = $this->klarnaManagerFactory->get($this->getConfiguration());
      $klarna_manager->cancelOrder($payment->getOrder());
      $payment->state = 'voided';
      $payment->save();
    }
    catch (\Exception $exception) {
      throw new PaymentGatewayException($exception->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, Price $amount = NULL) {
    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);
    // If not specified, refund the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $this->assertRefundAmount($payment, $amount);
    $order = $payment->getOrder();

    $klarna_manager = $this->klarnaManagerFactory->get($this->getConfiguration());
    try {
      $parameters = [
        'refunded_amount' => $this->toMinorUnits($amount),
      ];
      $klarna_manager->createRefund($order, $parameters);
    }
    catch (\Exception $exception) {
      throw new PaymentGatewayException($exception->getMessage());
    }
    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($amount);
    if ($new_refunded_amount->lessThan($payment->getAmount())) {
      $payment->state = 'partially_refunded';
    }
    else {
      $payment->state = 'refunded';
    }

    $payment->setRefundedAmount($new_refunded_amount);
    $payment->save();
  }

  /**
   * Converts an amount in "minor unit" to a decimal amount.
   *
   * For example, 999 USD becomes 9.99.
   *
   * @param mixed $amount
   *   The amount in minor unit.
   * @param string $currency_code
   *   The currency code.
   *
   * @return \Drupal\commerce_price\Price
   *   The decimal price.
   */
  protected function fromMinorUnits($amount, $currency_code) {
    $currency_storage = $this->entityTypeManager->getStorage('commerce_currency');
    /** @var \Drupal\commerce_price\Entity\CurrencyInterface $currency */
    $currency = $currency_storage->load($currency_code);
    $fraction_digits = $currency->getFractionDigits();

    if ($fraction_digits > 0) {
      $amount = Calculator::divide((string) $amount, pow(10, $fraction_digits), $fraction_digits);
    }

    return new Price((string) $amount, $currency_code);
  }

  /**
   * Populate the given profile with the given Klarna address.
   *
   * @param \Drupal\profile\Entity\ProfileInterface $profile
   *   The profile to populate.
   * @param array $address
   *   The Klarna address.
   */
  protected function populateProfile(ProfileInterface $profile, array $address) {
    $mapping = [
      'organization_name' => 'organization',
      'street_address' => 'address_line1',
      'street_address2' => 'address_line2',
      'city' => 'locality',
      'region' => 'administrative_area',
      'postal_code' => 'postal_code',
      'country' => 'country_code',
      'given_name' => 'given_name',
      'family_name' => 'family_name',
    ];
    foreach ($address as $key => $value) {
      if (!isset($mapping[$key])) {
        continue;
      }
      $value = $key === 'country' ? strtoupper($value) : $value;
      $profile->address->{$mapping[$key]} = $value;
    }
  }

}
